/* $Id: sys.h,v 1.88 2004-10-21 14:58:19 cartho Exp $ */
/*------------------------------------------------------------------------*/

#ifndef _JNUKE_sys_h_INCLUDED
#define _JNUKE_sys_h_INCLUDED
#define DIR_SEP "/"
/* directory separator, "/" is accepted on any decent operating systems */

#define PATH_SEP ":"
/* path seperator, i.e. in a CLASSPATH environment variable */

/*------------------------------------------------------------------------*/

union JNuke4bytes
{
  unsigned char byte[4];
};

union JNuke8bytes
{
  unsigned char byte[8];
};

struct JNukeMem
{
  int bytes;			/* current number of bytes allocated */
  int max;			/* maximal number of bytes allocated */
  int idx;			/* next idx of allocated MemBucket */
#ifdef JNUKE_TEST
  int force_out_of_memory;
#endif
};
typedef struct JNukeMem JNukeMem;
typedef struct JNukeType JNukeType;	/* Common object type descriptor */
typedef struct JNukeObj JNukeObj;	/* Common object reference */
typedef struct UCSString UCSString;	/* Universal Character Set string */
JNukeMem *JNukeMem_new (void);
int JNukeMem_delete (JNukeMem *);	/* returns leaked bytes */

/*------------------------------------------------------------------------*/
/* Returns a platform independent hash value, which is generated by sequence
 * numbers of allocated memory blocks.  Since the sequence numbers may wrap
 * around this value may not be unique.
 */
int JNuke_hash (JNukeMem *, void *);

/*------------------------------------------------------------------------*/
/* We can set the internally sequence number of future memory blocks
 * explicitly.  This can make the sequence numbers, which are hash values,
 * of future memory blocks, independent of the number of previously allocate
 * memory blocks.
 */
void JNuke_set_hash_idx (JNukeMem * mem, int new_idx);

/*------------------------------------------------------------------------*/

int JNuke_bytes_allocated (JNukeMem *);
int JNuke_max_bytes_allocated (JNukeMem *);

/*------------------------------------------------------------------------*/
/* In order to be able to calculate the number of bytes allocated we require
 * the user to specify the size of all freed data.  In debugging mode the
 * program checks for invalid sizes as second argument to 'JNuke_free'.
 */
void *JNuke_malloc (JNukeMem *, int size);
void JNuke_free (JNukeMem *, void *, int size);
void *JNuke_realloc (JNukeMem *, void *ptr, int old_size, int new_size);

/*------------------------------------------------------------------------*/
/* Comparison function for object values. Use for setComparator.
 */
int JNuke_hash_obj (void *);
int JNuke_cmp_obj (void *, void *);

/*------------------------------------------------------------------------*/
/* Hash and comparison functions for 'int' values.
 */
int JNuke_hash_int (void *);
int JNuke_cmp_int (void *, void *);
char *JNuke_printf_int (JNukeMem *, const void *);

/*------------------------------------------------------------------------*/
/* Default hash and comparison functions for pointers.
 */
int JNuke_hash_pointer (void *);
int JNuke_cmp_pointer (void *, void *);
char *JNuke_printf_pointer (JNukeMem * mem, const void *);
void *JNuke_scanf_pointer (const char *);

/*------------------------------------------------------------------------*/
/* This is the interface for all fields and methods that have to be
 * implemented for each object. Instances of the same type should
 * share only one such type descriptor.
 *
 * structs JNukeType and JNukeObj should only be used by the class
 * implementing it, not by clients (they should use
 * JNukeObj_<function> instead) 
 */
struct JNukeType
{
  /* name of object type as a string */
  const char *name;

  /* returns a deep copy of the instance */
  JNukeObj *(*clone) (const JNukeObj *);

  /* deletes instance of object */
  void (*delete) (JNukeObj *);

  /* returns comparison result as in strcmp(3)
   * result only needs to be defined for objects of equal type;
   * otherwise, pointer value is used (no true polymorphism) 
   */
  int (*cmp) (const JNukeObj *, const JNukeObj *);

  /* returns hash code of object */
  int (*hash) (const JNukeObj *);

  /* returns new string with string representation of instance */
  char *(*toString) (const JNukeObj *);

  /* clears object data (only used for containers: calls delete for
   * each element. "Deep" (recursive) clears have to be done manually,
   * but occur rarely in practice. */
  void (*clear) (JNukeObj *);

  /* allow for subtyping */
  void *subtype;
};

/*------------------------------------------------------------------------*/
/* Generic object instance reference 
 */
struct JNukeObj
{
  JNukeType *type;		/* type information */
  JNukeMem *mem;		/* memory manager */
  void *obj;			/* object instance */
};

/*------------------------------------------------------------------------*/

/* Generic object functions, providing encapsulation from implementation*/

int JNukeObj_isType (const JNukeObj *, JNukeType);
const char *JNukeObj_typeName (const JNukeObj *);
JNukeObj *JNukeObj_clone (const JNukeObj *);
void JNukeObj_delete (JNukeObj *);
int JNukeObj_cmp (const JNukeObj *, const JNukeObj *);
int JNukeObj_hash (const JNukeObj *);
char *JNukeObj_toString (const JNukeObj *);
int JNukeObj_isContainer (const JNukeObj *);
  /* returns true iff object is a container and implements a clear
     method */
void JNukeObj_clear (JNukeObj *);

/*------------------------------------------------------------------------*/
/* If you use this 'strdup' make sure that you remember how long the string
 * is when deleting it with 'JNuke_free' later on.
 */
char *JNuke_strdup (JNukeMem *, const char *str);
char *JNuke_strdupN (JNukeMem *, const char *str, int size);
/* strdupN reserves n + 1 bytes, adding a \0 at the end of the string! */

/*------------------------------------------------------------------------*/
/* Return the time spent in this process in seconds.
 */
double JNuke_process_time (void);

/*------------------------------------------------------------------------*/
/* Return '1' iff the name given as second argument is an existing file or
 * directory respectively.  Otherwise '0' is returned.
 */
int JNuke_is_file (const char *file_name);
int JNuke_is_directory (const char *file_name);

/*------------------------------------------------------------------------*/
/* Wrapper object for integers */
extern JNukeType JNukeIntType;
JNukeObj *JNukeInt_new (JNukeMem *);
void JNukeInt_set (JNukeObj *, int);
int JNukeInt_value (const JNukeObj *);

/*------------------------------------------------------------------------*/
/* Wrapper object for pointers */
extern JNukeType JNukePtrType;
JNukeObj *JNukePtr_new (JNukeMem *);
void JNukePtr_set (JNukeObj *, void *);
void *JNukePtr_value (const JNukeObj *);

/*------------------------------------------------------------------------*/
/* Wrapper object for floats */

extern JNukeType JNukeFloatType;
JNukeObj *JNukeFloat_new (JNukeMem *);
void JNukeFloat_set (JNukeObj *, JNukeJFloat);
JNukeJFloat JNukeFloat_value (const JNukeObj *);
union JNuke4bytes JNukeFloat_toIEEE754 (JNukeObj *);
int JNukeFloat_isNaN (const JNukeObj * this);
int JNukeFloat_isInf (const JNukeObj * this);
JNukeJFloat JNukeFloat_unpack (unsigned int bytes);

/*------------------------------------------------------------------------*/
/* Wrapper object for longs */
extern JNukeType JNukeLongType;
JNukeObj *JNukeLong_new (JNukeMem *);
void JNukeLong_set (JNukeObj *, JNukeInt8);
JNukeInt8 JNukeLong_value (const JNukeObj *);

/*------------------------------------------------------------------------*/
/* Wrapper object for doubles */
extern JNukeType JNukeDoubleType;
JNukeObj *JNukeDouble_new (JNukeMem *);
void JNukeDouble_set (JNukeObj *, JNukeJDouble);
JNukeJDouble JNukeDouble_value (const JNukeObj *);
union JNuke8bytes JNukeDouble_toIEEE754 (JNukeObj * this);
int JNukeDouble_isNaN (const JNukeObj * this);
int JNukeDouble_isInf (const JNukeObj * this);
JNukeJDouble JNukeDouble_unpack (unsigned long long bytes);

/*------------------------------------------------------------------------*/
/* Unicode support for UCSString */

extern JNukeType UCSStringType;
JNukeObj *UCSString_new (JNukeMem *, const char *);
/* New UCS string from ASCII or UTF8, but not ISO-latin! */
JNukeObj *UCSString_newN (JNukeMem *, const char *, int n);
/* New UCS string from non-\0 terminated string (length must be given) */

const char *UCSString_toUTF8 (const JNukeObj *);
/* returns char * content, in UTF8 format */
const wchar_t *UCSString_toUCS (const JNukeObj * this);
/* returns wchar_t * content in UCS format */
char *UCSString_deleteBuffer (JNukeObj *);
/* Deletes object but leaves UTF8 char * buffer. Returns that buffer */

int UCSString_UTF8length (const JNukeObj *);
int UCSString_length (const JNukeObj *);
int UCSString_append (const JNukeObj *, const char *);
int UCSString_appendN (const JNukeObj *, const char *, int);
/* both append functions return the number of characters appended,
   i.e., strlen of the second argument */

JNukeObj *UCSString_dup (const JNukeObj *);
void UCSString_tr (JNukeObj *, char, char);
/* replace all occurrences of first char with second char */
/* first char must not be '\0' and different from second char */

/* Return '1' iff [from1, to1] is a proper subrange of [from2, to2] */
int JNuke_is_proper_subrange (int from1, int to1, int from2, int to2);

/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/
/* TimeStamp */

extern JNukeType JNukeTimeStampType;

JNukeObj *JNukeTimeStamp_new (JNukeMem * mem);

void JNukeTimeStamp_setDate (JNukeObj * this, const int d, const int m,
			     const int y);
void JNukeTimeStamp_setTime (JNukeObj * this, const int h, const int m,
			     const int s);

int JNukeTimeStamp_getYear (const JNukeObj * this);
int JNukeTimeStamp_getMonth (const JNukeObj * this);
int JNukeTimeStamp_getDay (const JNukeObj * this);
int JNukeTimeStamp_getHour (const JNukeObj * this);
int JNukeTimeStamp_getMinute (const JNukeObj * this);
int JNukeTimeStamp_getSecond (const JNukeObj * this);
/*------------------------------------------------------------------------*/

#endif
